@@ -236,6 +236,7 @@ static OP *pp_argelems_named(pTHX)
   AV *defav = GvAV(PL_defgv);
 
   HV *slurpy_hv = NULL;
+  AV *slurpy_av = NULL;
 
   if(PL_op->op_targ) {
     /* We have a slurpy of some kind */
@@ -246,17 +247,23 @@ static OP *pp_argelems_named(pTHX)
       assert(SvTYPE(slurpy_hv) == SVt_PVHV);
       assert(HvKEYS(slurpy_hv) == 0);
     }
+    else if(PL_op->op_private & OPpARGELEM_AV) {
+      slurpy_av = (AV *)PAD_SVl(PL_op->op_targ);
+      assert(SvTYPE(slurpy_av) == SVt_PVAV);
+      assert(av_count(slurpy_av) == 0);
+    }
   }
 
   UV argix = aux->start_argix;
   UV argc  = av_count(defav);
 
+  U32 parami;
   UV n_params = aux->n_params;
 
   /* Before we process the incoming args we need to prepare *all* the param
    * variable pad slots.
    */
-  for(U32 parami = 0; parami < n_params; parami++) {
+  for(parami = 0; parami < n_params; parami++) {
     struct ArgElemsNamedParam *param = &aux->params[parami];
 
     SV **padentry = &PAD_SVl(param->padix);
@@ -274,8 +281,10 @@ static OP *pp_argelems_named(pTHX)
 
   while(argix < argc) {
     /* TODO: do we need av_fetch or can we cheat around it? */
-    SV *name = *av_fetch(defav, argix++, 0);
-    SV *val  = argix < argc ? *av_fetch(defav, argix++, 0) : &PL_sv_undef;
+    SV *name = *av_fetch(defav, argix, 0);
+    argix++;
+    SV *val  = argix < argc ? *av_fetch(defav, argix, 0) : &PL_sv_undef;
+    argix++;
 
     STRLEN namelen;
     const char *namepv = SvPV(name, namelen);
@@ -285,12 +294,15 @@ static OP *pp_argelems_named(pTHX)
 
     PADOFFSET param_padix = 0;
 
-    /* TODO: We can get better performance at runtime by binary searching
-     * for a good starting index. In practice it probably only actually saves
-     * measurable time once we start to get tens or even hundreds of named
-     * params. Is that likely?
+    /* In theory we would get better performance at runtime by binary
+     * searching for a good starting index. In practice only actually starts
+     * saving measurable time once we start to get to literally hundreds of
+     * named parameters. This simple linear search is actually very quick per
+     * rejected element.
+     * If your perl function wants to declare hundreds of different named
+     * parameters you probably want to rethink your strategy. ;)
      */
-    for(U32 parami = 0; parami < n_params; parami++) {
+    for(parami = 0; parami < n_params; parami++) {
       struct ArgElemsNamedParam *param = &aux->params[parami];
 
       /* Since the params are stored in hash key order, if we are already
@@ -324,6 +336,11 @@ static OP *pp_argelems_named(pTHX)
     else if(slurpy_hv) {
       hv_store_ent(slurpy_hv, name, newSVsv(val), 0);
     }
+    else if(slurpy_av) {
+      av_push(slurpy_av, newSVsv(name));
+      if(argix <= argc)
+        av_push(slurpy_av, newSVsv(val));
+    }
     else {
       if(!unrecognised_keynames) {
         unrecognised_keynames = newSVpvn("", 0);
@@ -343,7 +360,10 @@ static OP *pp_argelems_named(pTHX)
       SVfARG(unrecognised_keynames), SVfARG(S_find_runcv_name(aTHX)));
   }
 
-  for(U32 parami = 0; parami < n_params; parami++) {
+  SV *missing_keynames = NULL;
+  UV n_missing = 0;
+
+  for(parami = 0; parami < n_params; parami++) {
     struct ArgElemsNamedParam *param = &aux->params[parami];
     SV *targ = PAD_SVl(param->padix);
 
@@ -352,8 +372,21 @@ static OP *pp_argelems_named(pTHX)
     if(!(param->flags & NAMEDPARAMf_REQUIRED))
       continue;
 
-    croak_from_caller("Missing argument '%s' for subroutine %" SVf,
-      param->namepv, SVfARG(S_find_runcv_name(aTHX)));
+    if(!missing_keynames) {
+      missing_keynames = newSVpvn("", 0);
+      SAVEFREESV(missing_keynames);
+    }
+
+    if(SvCUR(missing_keynames))
+      sv_catpvs(missing_keynames, ", ");
+    sv_catpvf(missing_keynames, "'%s'", param->namepv);
+    n_missing++;
+  }
+
+  if(n_missing) {
+    croak_from_caller("Missing %s %" SVf " for subroutine %" SVf,
+      n_missing > 1 ? "arguments" : "argument",
+      SVfARG(missing_keynames), SVfARG(S_find_runcv_name(aTHX)));
   }
 
   return NORMAL;
@@ -366,7 +399,9 @@ static void opdump_argelems_named(pTHX_ const OP *o, struct Perl_OpDumpContext *
 
   opdump_printf(ctx, "START_ARGIX = %" UVuf "\n", aux->start_argix);
   opdump_printf(ctx, "PARAMS = (%" UVuf ")\n", aux->n_params);
-  for(U32 parami = 0; parami < aux->n_params; parami++) {
+
+  U32 parami;
+  for(parami = 0; parami < aux->n_params; parami++) {
     struct ArgElemsNamedParam *param = &aux->params[parami];
 
     opdump_printf(ctx, "  [%d] = {.name=\"%s\", .namehash=%u .padix=%u, .flags=(",
@@ -723,10 +758,14 @@ endofelems:
     "The signatures feature is experimental");
 #endif
 
+  char sig_slurpy = parser->sig_slurpy;
+  if(!sig_slurpy && ctx.named_details)
+    sig_slurpy = '%';
+
   UNOP_AUX_item *aux = make_argcheck_aux(
     parser->sig_elems,
     parser->sig_optelems,
-    ctx.named_details ? '%' : parser->sig_slurpy);
+    sig_slurpy);
 
   OP *checkop = newUNOP_AUX(OP_ARGCHECK, 0, NULL, aux);
 
